@@ -3,15 +3,11 @@ package Method::Signatures;
 use strict;
 use warnings;
 
+use base 'Devel::Declare::MethodInstaller::Simple';
 use Method::Signatures::Parser;
+use Data::Alias;
 
-use Devel::Declare ();
-use Data::Alias ();
-use Readonly;
-use Scope::Guard;
-use Sub::Name;
-
-our $VERSION = '20081021.1911';
+our $VERSION = '20110216.1153_01';
 
 our $DEBUG = $ENV{METHOD_SIGNATURES_DEBUG} || 0;
 
@@ -25,34 +21,41 @@ sub DEBUG {
 
 =head1 NAME
 
-Method::Signatures - method declarations with signatures and no source filter
+Method::Signatures - method and function declarations with signatures and no source filter
 
 =head1 SYNOPSIS
 
     package Foo;
-    
+
     use Method::Signatures;
-    
+
     method new (%args) {
         return bless {%args}, $self;
     }
-    
+
     method get ($key) {
         return $self->{$key};
     }
-    
+
     method set ($key, $val) {
         return $self->{$key} = $val;
     }
 
+    func hello($greeting, $place) {
+        print "$greeting, $place!\n";
+    }
+
 =head1 DESCRIPTION
 
-This is B<ALPHA SOFTWARE> which relies on B<YET MORE ALPHA SOFTWARE>.
-Use at your own risk.  Features may change.
+Provides two new keywords, C<func> and C<method> so you can write subroutines with signatures instead of having to spell out C<my $self = shift; my($thing) = @_>
 
-Provides a proper method keyword, like "sub" but specificly for making
-methods.  It will automatically provide the invocant as C<$self>.  No
-more C<my $self = shift>.
+C<func> is like C<sub> but takes a signature where the prototype would
+normally go.  This takes the place of C<my($foo, $bar) = @_> and does
+a whole lot more.
+
+C<method> is like C<func> but specificly for making methods.  It will
+automatically provide the invocant as C<$self>.  No more C<my $self =
+shift>.
 
 Also allows signatures, very similar to Perl 6 signatures.
 
@@ -61,6 +64,22 @@ And it does all this with B<no source filters>.
 
 =head2 Signature syntax
 
+    func echo($message) {
+        print "$message\n";
+    }
+
+is equivalent to:
+
+    sub echo {
+        my($message) = @_;
+        print "$message\n";
+    }
+
+except the original line numbering is preserved and the arguments are
+checked to make sure they match the signature.
+
+Similarly
+
     method foo($bar, $baz) {
         $self->wibble($bar, $baz);
     }
@@ -73,14 +92,6 @@ is equivalent to:
         $self->wibble($bar, $baz);
     }
 
-except the original line numbering is preserved.
-
-No checks are made that the arguments being passed in match the
-signature.
-
-Future releases will add extensively to the signature syntax probably
-along the lines of Perl 6.
-
 
 =head3 C<@_>
 
@@ -300,7 +311,7 @@ You can alias using the "alias" trait.
 
 =head3 Can't use positional params as named params
 
-Perl 6 allows you to use any parameter as a named paramter.  Perl 5
+Perl 6 allows you to use any parameter as a named parameter.  Perl 5
 lacks the named parameter disambiguating syntax so it is not allowed.
 
 =head3 Addition of the C<\@foo> reference alias prototype
@@ -323,17 +334,32 @@ sub import {
     my $arg = shift;
     $DEBUG = 1 if defined $arg and $arg eq ':DEBUG';
 
-    Devel::Declare->setup_for(
-        $caller,
-        { method => { const => \&parser } }
+    $class->install_methodhandler(
+        into            => $caller,
+        name            => 'method',
+        invocant        => '$self'
+    );
+
+    $class->install_methodhandler(
+        into            => $caller,
+        name            => 'func',
     );
 
     DEBUG("import for $caller done\n");
+}
 
-    # I don't really understand why we need to declare method
-    # in the caller's namespace.
-    no strict 'refs';
-    *{$caller.'::method'} = sub (&) {};
+
+sub code_for {
+    my($self, $name) = @_;
+
+    my $code = $self->SUPER::code_for($name);
+
+    if( defined $name ) {
+        require Devel::BeginLift;
+        Devel::BeginLift->setup_for_cv($code);
+    }
+
+    return $code;
 }
 
 
@@ -343,29 +369,65 @@ sub _strip_ws {
 }
 
 
-sub make_proto_unwrap {
-    my ($proto) = @_;
-    $proto ||= '';
+# Overriden method from D::D::MS
+sub parse_proto {
+    my $self = shift;
+    return $self->parse_signature( proto => shift, invocant => $self->{invocant} );
+}
 
-    _strip_ws($proto);
 
-    my @protos = split_proto($proto);
+# Parse a signature
+sub parse_signature {
+    my $self = shift;
+    my %args = @_;
+    my @protos = $self->_split_proto($args{proto} || []);
+    my $signature = $args{signature} || {};
 
-    my %signature;
-    $signature{invocant} = '$self';
-    if( @protos ) {
-        $signature{invocant} = $1 if $protos[0] =~ s{^(\S+?):\s*}{};
-        shift @protos unless $protos[0] =~ /\S/;
+    # Special case for methods, they will pass in an invocant to use as the default
+    if( $signature->{invocant} = $args{invocant} ) {
+        if( @protos ) {
+            $signature->{invocant} = $1 if $protos[0] =~ s{^(\S+?):\s*}{};
+            shift @protos unless $protos[0] =~ /\S/;
+        }
     }
 
-    $signature{named}      = [];
-    $signature{positional} = [];
-    $signature{overall}    = {
+    return $self->parse_func( proto => \@protos, signature => $signature );
+}
+
+
+sub _split_proto {
+    my $self = shift;
+    my $proto = shift;
+
+    my @protos;
+    if( ref $proto ) {
+        @protos = @$proto;
+    }
+    else {
+        _strip_ws($proto);
+        @protos = split_proto($proto);
+    }
+
+    return @protos;
+}
+
+
+# Parse a subroutine signature
+sub parse_func {
+    my $self = shift;
+    my %args = @_;
+    my @protos = $self->_split_proto($args{proto} || []);
+    my $signature = $args{signature} || {};
+
+    $signature->{named}      = [];
+    $signature->{positional} = [];
+    $signature->{overall}    = {
         has_optional            => 0,
         has_optional_positional => 0,
         has_named               => 0,
         has_positional          => 0,
-        has_invocant            => $signature{invocant} ? 1 : 0,
+        has_invocant            => $signature->{invocant} ? 1 : 0,
+        num_slurpy              => 0
     };
 
     my $idx = 0;
@@ -395,28 +457,30 @@ sub make_proto_unwrap {
         $sig->{sigil}       = $sigil;
         $sig->{name}        = $name;
         $sig->{var}         = $sigil . $name;
+        $sig->{is_slurpy}   = ($sigil =~ /^[%@]$/ and !$sig->{is_ref_alias});
 
-        check_signature($sig, \%signature);
+        check_signature($sig, $signature);
 
         if( $sig->{named} ) {
-            push @{$signature{named}}, $sig;
+            push @{$signature->{named}}, $sig;
         }
         else {
-            push @{$signature{positional}}, $sig;
+            push @{$signature->{positional}}, $sig;
         }
 
-        $signature{has_optional}++              if $sig->{is_optional};
-        $signature{has_named}++                 if $sig->{named};
-        $signature{has_positional}++            if !$sig->{named};
-        $signature{has_optional_positional}++   if $sig->{is_optional} and !$sig->{named};
+        my $overall = $signature->{overall};
+        $overall->{has_optional}++              if $sig->{is_optional};
+        $overall->{has_named}++                 if $sig->{named};
+        $overall->{has_positional}++            if !$sig->{named};
+        $overall->{has_optional_positional}++   if $sig->{is_optional} and !$sig->{named};
+        $overall->{num_slurpy}++                if $sig->{is_slurpy};
 
         DEBUG( "sig: ", $sig );
     }
 
     # Then turn it into Perl code
-    my $inject = inject_from_signature(\%signature);
+    my $inject = inject_from_signature($signature);
     DEBUG( "inject: $inject\n" );
-
     return $inject;
 }
 
@@ -424,14 +488,17 @@ sub make_proto_unwrap {
 sub check_signature {
     my($sig, $signature) = @_;
 
+    die("signature can only have one slurpy parameter") if
+      $sig->{is_slurpy} and $signature->{overall}{num_slurpy} >= 1;
+
     if( $sig->{named} ) {
-        if( $signature->{has_optional_positional} ) {
+        if( $signature->{overall}{has_optional_positional} ) {
             my $pos_var = $signature->{positional}[-1]{var};
             die("named parameter $sig->{var} mixed with optional positional $pos_var\n");
         }
     }
     else {
-        if( $signature->{has_named} ) {
+        if( $signature->{overall}{has_named} ) {
             my $named_var = $signature->{named}[-1]{var};
             die("positional parameter $sig->{var} after named param $named_var\n");
         }
@@ -444,7 +511,7 @@ sub inject_from_signature {
     my $signature = shift;
 
     my @code;
-    push @code, "my $signature->{invocant} = shift;";
+    push @code, "my $signature->{invocant} = shift;" if $signature->{invocant};
 
     for my $sig (@{$signature->{positional}}) {
         push @code, inject_for_sig($sig);
@@ -459,19 +526,17 @@ sub inject_from_signature {
         push @code, inject_for_sig($sig);
     }
 
-    push @code, 'Method::Signatures::named_param_check(\%args);' if $signature->{has_named};
+    push @code, 'Method::Signatures::named_param_error(\%args) if %args;' if $signature->{overall}{has_named};
 
     # All on one line.
     return join ' ', @code;
 }
 
 
-sub named_param_check {
+sub named_param_error {
     my $args = shift;
     my @keys = keys %$args;
 
-    return 1 unless @keys;
-
     signature_error("does not take @keys as named argument(s)");
 }
 
@@ -516,7 +581,8 @@ sub inject_for_sig {
     }
     # Handle "is ro"
     elsif ( $sig->{traits}{ro} ) {
-        push @code, "Readonly::Readonly $lhs => $rhs;";
+        require Const::Fast;
+        push @code, "Const::Fast::const( $lhs => $rhs );";
     } else {
         push @code, "$lhs = $rhs;";
     }
@@ -539,160 +605,6 @@ sub required_arg {
 }
 
 
-# Stolen from Devel::Declare's t/method-no-semi.t
-{
-    our ($Declarator, $Offset);
-
-    sub skip_declarator {
-        $Offset += Devel::Declare::toke_move_past_token($Offset);
-    }
-
-    sub skipspace {
-        $Offset += Devel::Declare::toke_skipspace($Offset);
-    }
-
-    sub strip_name {
-        skipspace;
-        if (my $len = Devel::Declare::toke_scan_word($Offset, 1)) {
-            my $linestr = Devel::Declare::get_linestr();
-            my $name = substr($linestr, $Offset, $len);
-            substr($linestr, $Offset, $len) = '';
-            Devel::Declare::set_linestr($linestr);
-            return $name;
-        }
-        return;
-    }
-
-    sub strip_proto {
-        skipspace;
-    
-        my $linestr = Devel::Declare::get_linestr();
-        DEBUG( "strip_proto/\$linestr: $linestr\n" );
-        if (substr($linestr, $Offset, 1) eq '(') {
-            my $length = Devel::Declare::toke_scan_str($Offset);
-            my $proto = Devel::Declare::get_lex_stuff();
-            Devel::Declare::clear_lex_stuff();
-            if( $length < 0 ) {
-                # Need to scan ahead more
-                $linestr .= Devel::Declare::get_linestr();
-                $length = length($linestr) - rindex($linestr, ")") + 1;
-            }
-            else {
-                $linestr = Devel::Declare::get_linestr();
-            }
-
-            DEBUG("strip_proto/Offset: $Offset, length: $length, linestr, '$linestr'\n");
-            substr($linestr, $Offset, $length) = '';
-            DEBUG("strip_proto/after substr: linestr, '$linestr'\n");
-
-            Devel::Declare::set_linestr($linestr);
-
-            DEBUG( "strip_proto/\$proto: $proto\n" );
-
-            return $proto;
-        }
-        return;
-    }
-
-    sub shadow {
-        my $pack = Devel::Declare::get_curstash_name;
-        Devel::Declare::shadow_sub("${pack}::${Declarator}", $_[0]);
-    }
-
-    # Improved attributed parsing from MooseX::Method::Signatures
-    sub inject_if_block {
-        my $inject = shift;
-        skipspace;
-        my $linestr = Devel::Declare::get_linestr;
-
-        my $attrs   = '';
-
-        DEBUG("inject_if_block/\$linestr: $linestr\n");
-        if (substr($linestr, $Offset, 1) eq ':') {
-            while (substr($linestr, $Offset, 1) ne '{') {
-                if (substr($linestr, $Offset, 1) eq ':') {
-                    substr($linestr, $Offset, 1) = '';
-                    Devel::Declare::set_linestr($linestr);
-
-                    $attrs .= ' :';
-                }
-
-                skipspace;
-                $linestr = Devel::Declare::get_linestr();
-
-                if (my $len = Devel::Declare::toke_scan_word($Offset, 0)) {
-                    my $name = substr($linestr, $Offset, $len);
-                    substr($linestr, $Offset, $len) = '';
-                    Devel::Declare::set_linestr($linestr);
-
-                    $attrs .= " ${name}";
-
-                    if (substr($linestr, $Offset, 1) eq '(') {
-                        my $length = Devel::Declare::toke_scan_str($Offset);
-                        my $arg    = Devel::Declare::get_lex_stuff();
-                        Devel::Declare::clear_lex_stuff();
-                        $linestr = Devel::Declare::get_linestr();
-                        substr($linestr, $Offset, $length) = '';
-                        Devel::Declare::set_linestr($linestr);
-
-                        $attrs .= "(${arg})";
-                    }
-                }
-            }
-
-            $linestr = Devel::Declare::get_linestr();
-        }
-
-        if (substr($linestr, $Offset, 1) eq '{') {
-            substr($linestr, $Offset + 1, 0) = $inject;
-            substr($linestr, $Offset, 0) = "sub ${attrs}";
-            Devel::Declare::set_linestr($linestr);
-        }
-
-        DEBUG("inject_if_block done\n");
-    }
-
-    sub scope_injector_call {
-        return ' BEGIN { Method::Signatures::inject_scope }; ';
-    }
-
-    sub parser {
-        local ($Declarator, $Offset) = @_;
-        skip_declarator;
-        my $name = strip_name;
-        my $proto = strip_proto;
-        my $inject = make_proto_unwrap($proto);
-        if (defined $name) {
-            $inject = scope_injector_call().$inject;
-        }
-        inject_if_block($inject);
-        if (defined $name) {
-            $name = join('::', Devel::Declare::get_curstash_name(), $name)
-              unless ($name =~ /::/);
-            shadow(sub (&) {
-                no strict 'refs';
-                # So caller() gets the subroutine name
-                *{$name} = subname $name => shift;
-            });
-        } else {
-            shadow(sub (&) { shift });
-        }
-
-        DEBUG("parser done\n");
-    }
-
-    sub inject_scope {
-        $^H |= 0x120000;
-        $^H{DD_METHODHANDLERS} = Scope::Guard->new(sub {
-            my $linestr = Devel::Declare::get_linestr;
-            my $offset = Devel::Declare::get_linestr_offset;
-            substr($linestr, $offset, 0) = ';';
-            Devel::Declare::set_linestr($linestr);
-        });
-    }
-}
-
-
 =head1 PERFORMANCE
 
 There is no run-time performance penalty for using this module above
@@ -731,7 +643,7 @@ The display() method is equivalent to all this code.
       my $self = shift;
 
       croak('display() missing required argument $text') unless @_ > 0;
-      Readonly my $text = $_[0];
+      const my $text = $_[0];
 
       my(%args) = @_[1 .. $#_];
       my $justify = exists $args{justify} ? $args{justify} : 'left';
@@ -758,16 +670,8 @@ Please report bugs and leave feedback at
 E<lt>bug-Method-SignaturesE<gt> at E<lt>rt.cpan.orgE<gt>.  Or use the
 web interface at L<http://rt.cpan.org>.  Report early, report often.
 
-=head2 C<method> is not declared at compile time.
-
-Unlike declaring a C<sub>, C<method> currently does not happen at
-compile time.  This usually isn't a problem.  It may change, but it
-may be a good thing.
-
 =head2 Debugging
 
-This totally breaks the debugger.  Will have to wait on Devel::Declare fixes.
-
 You can see the Perl code Method::Signatures translates to by using B::Deparse.
 
 =head2 One liners
@@ -778,10 +682,10 @@ Devel::Declare.
 
 =head2 No source filter
 
-While this module does rely on the hairy black magic of
-L<Devel::Declare> and L<Data::Alias> it does not depend on a source
-filter.  As such, it doesn't try to parse and rewrite your source code
-and there should be no weird side effects.
+While this module does rely on the black magic of L<Devel::Declare> to
+access Perl's own parser, it does not depend on a source filter.  As
+such, it doesn't try to parse and rewrite your source code and there
+should be no weird side effects.
 
 Devel::Declare only effects compilation.  After that, it's a normal
 subroutine.  As such, for all that hairy magic, this module is
@@ -876,8 +780,8 @@ Method traits.
 Most of this module is based on or copied from hard work done by many
 other people.
 
-All the really scary parts are copied from or rely on Matt Trout's
-L<Devel::Declare> work.
+All the really scary parts are copied from or rely on Matt Trout's,
+Florian Ragwitz's and Rhesa Rozendaal's L<Devel::Declare> work.
 
 The prototype syntax is a slight adaptation of all the
 excellent work the Perl 6 folks have already done.
@@ -888,7 +792,8 @@ makes the subroutine names come out right in caller().
 
 And thanks to Florian Ragwitz for his parallel
 L<MooseX::Method::Signatures> module from which I borrow ideas and
-code.
+code and L<Devel::BeginLift> which lets the methods be declared
+at compile time.
 
 
 =head1 LICENSE
@@ -905,7 +810,13 @@ See F<http://www.perl.com/perl/misc/Artistic.html>
 
 =head1 SEE ALSO
 
-L<MooseX::Method::Signatures>, L<Perl6::Signature>, L<Sub::Signatures>, L<Perl6::Subs>
+L<MooseX::Method::Signatures> for a method keyword that works well with Moose.
+
+L<Perl6::Signature> for a more complete implementation of Perl 6 signatures.
+
+L<Method::Signatures::Simple> for a more basic version of what Method::Signatures provides.
+
+L<signatures> for C<sub> with signatures.
 
 Perl 6 subroutine parameters and arguments -  L<http://perlcabal.org/syn/S06.html#Parameters_and_arguments>
 
